"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.styleLoader = exports.convertLessError = exports.transformLess = void 0;
const path_1 = __importDefault(require("path"));
const lightningcss_1 = require("lightningcss");
const promises_1 = require("fs/promises");
const query_string_1 = __importDefault(require("query-string"));
const deepmerge_1 = __importDefault(require("deepmerge"));
const colors_1 = __importDefault(require("colors"));
const transform_less_1 = require("./transform-less");
Object.defineProperty(exports, "transformLess", { enumerable: true, get: function () { return transform_less_1.transformLess; } });
const utils_1 = require("./utils");
const less_utils_1 = require("./less-utils");
Object.defineProperty(exports, "convertLessError", { enumerable: true, get: function () { return less_utils_1.convertLessError; } });
const transform_sass_1 = require("./transform-sass");
const sass_utils_1 = require("./sass-utils");
colors_1.default.enable();
const defaultOptions = {
    filter: /\.(css|scss|sass|less)(\?.*)?$/,
    cssModules: { pattern: '[local]__[hash]' },
    browsers: '> 0.25%, not dead',
};
const styleLoader = (options = {}) => {
    const opts = (0, deepmerge_1.default)(defaultOptions, options);
    const targets = (0, utils_1.generateTargets)(opts.browsers);
    const allNamespaces = Array.from(new Set(['file'].concat(opts.namespace || [])));
    const getLogger = (build) => {
        const { logLevel } = build.initialOptions;
        if (logLevel === 'debug' || logLevel === 'verbose') {
            return (...args) => {
                console.log(`[esbuild-style-loader]`.magenta.bold, ...args);
            };
        }
        return () => void 0;
    };
    return {
        name: 'style-loader',
        setup(build) {
            const buildOptions = build.initialOptions;
            const logger = getLogger(build);
            const cwd = process.cwd();
            const styleTransform = (filePath) => __awaiter(this, void 0, void 0, function* () {
                const extname = path_1.default.extname(filePath);
                if (extname === '.less') {
                    return yield (0, transform_less_1.transformLess)(filePath, {
                        sourcemap: !!buildOptions.sourcemap,
                        alias: buildOptions.alias,
                    }).catch((error) => {
                        logger(`transform less error: ${filePath}`.red.bold);
                        logger(error);
                        throw (0, less_utils_1.convertLessError)(error);
                    });
                }
                else if (extname === '.styl') {
                    // TODO: support stylus
                    throw new Error('stylus is not supported yet');
                }
                else if (extname === '.scss' || extname === '.sass') {
                    return yield (0, transform_sass_1.transformSass)(filePath, {
                        sourcemap: !!buildOptions.sourcemap,
                        alias: buildOptions.alias,
                    }).catch((error) => {
                        logger(`transform sass error: ${filePath}`.red.bold);
                        logger(error);
                        throw (0, sass_utils_1.convertScssError)(error, filePath);
                    });
                }
                else {
                    const code = yield (0, promises_1.readFile)(filePath, 'utf-8');
                    return { css: code, map: '' };
                }
            });
            const handleResolve = (args) => __awaiter(this, void 0, void 0, function* () {
                const { path: fullPath, query } = yield (0, utils_1.resolvePath)(args, build);
                return {
                    path: fullPath,
                    namespace: 'style-loader',
                    pluginData: {
                        rawPath: args.path,
                        query: query_string_1.default.parse(query),
                        resolveDir: args.resolveDir,
                    },
                };
            });
            /**
             * you can use `namespace` to control the order of the plugins
             */
            allNamespaces.forEach((namespace) => {
                build.onResolve({ filter: opts.filter, namespace }, handleResolve);
            });
            build.onLoad({ filter: /.*/, namespace: 'style-loader' }, (args) => __awaiter(this, void 0, void 0, function* () {
                let cssContent;
                let cssSourceMap;
                let watchImports = [];
                const pluginData = args.pluginData;
                let entryContent = (0, utils_1.cssExportsToJs)({}, pluginData.rawPath);
                // enable css modules
                // 1. if the file name contains `.modules.` or `.module.`
                // 2. if the query contains `modules`
                const enableCssModules = /\.modules?\.(css|less|sass|scss)/.test(args.path) || 'modules' in pluginData.query;
                let result;
                try {
                    const t = Date.now();
                    result = yield styleTransform(args.path);
                    logger(`Compile`, path_1.default.relative(cwd, args.path).blue.underline, `in ${Date.now() - t}ms`);
                    cssContent = result.css;
                    cssSourceMap = result.map;
                    watchImports = result.imports;
                }
                catch (error) {
                    return {
                        errors: [error],
                        resolveDir: path_1.default.dirname(args.path),
                    };
                }
                let transformResult;
                try {
                    const t = Date.now();
                    transformResult = (0, lightningcss_1.transform)({
                        targets: targets,
                        inputSourceMap: cssSourceMap,
                        sourceMap: true,
                        filename: args.path,
                        cssModules: enableCssModules ? opts.cssModules : false,
                        code: Buffer.from(cssContent),
                    });
                    logger(`Transform css in ${Date.now() - t}ms`);
                }
                catch (error) {
                    logger(`Transform css error: ${args.path}`.red.bold);
                    logger(error);
                    const { loc, fileName, source } = error;
                    const lines = source.split('\n');
                    const lineText = lines[loc.line - 1];
                    return {
                        errors: [
                            {
                                text: error.message,
                                location: {
                                    file: (0, utils_1.replaceExtension)(fileName, '.css'),
                                    line: loc.line,
                                    column: loc.column,
                                    lineText,
                                },
                            },
                        ],
                        resolveDir: path_1.default.dirname(args.path),
                    };
                }
                const { code, map, exports } = transformResult;
                if (buildOptions.sourcemap && map) {
                    cssContent = (0, utils_1.codeWithSourceMap)(code.toString(), map.toString());
                }
                else {
                    cssContent = code.toString();
                }
                if (exports) {
                    entryContent = (0, utils_1.cssExportsToJs)(exports, pluginData.rawPath);
                }
                return {
                    contents: entryContent,
                    loader: 'js',
                    pluginName: 'style-loader',
                    resolveDir: pluginData.resolveDir,
                    watchFiles: watchImports,
                    pluginData: Object.assign(Object.assign({}, pluginData), { cssContent }),
                    warnings: result.warnings,
                };
            }));
            build.onResolve({ filter: opts.filter, namespace: 'style-loader' }, (args) => __awaiter(this, void 0, void 0, function* () {
                const parsedPath = (0, utils_1.parsePath)(args.path);
                const result = yield build.resolve(parsedPath.path, {
                    kind: args.kind,
                    resolveDir: args.resolveDir,
                    importer: args.importer,
                });
                if (result.errors.length) {
                    return {
                        errors: result.errors,
                        resolveDir: path_1.default.dirname(args.path),
                    };
                }
                return {
                    path: result.path,
                    namespace: 'css-loader',
                    pluginData: args.pluginData,
                };
            }));
            build.onLoad({ filter: /.*/, namespace: 'css-loader' }, (args) => __awaiter(this, void 0, void 0, function* () {
                const pluginData = args.pluginData;
                const cssContent = pluginData.cssContent;
                return {
                    contents: cssContent,
                    loader: 'css',
                    resolveDir: path_1.default.dirname(args.path),
                };
            }));
            build.onResolve({ filter: /^\//, namespace: 'css-loader' }, (args) => __awaiter(this, void 0, void 0, function* () {
                if (opts.publicPath) {
                    // absolute files base on publicPath
                    return { path: path_1.default.join(opts.publicPath, '.' + args.path) };
                }
                return { external: true };
            }));
        },
    };
};
exports.styleLoader = styleLoader;
